#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <genht/hash.h>
#include <genht/htip.h>
#include "global_typedefs.h"
#include "htpxid.h"
#include "backend.h"
#include "display.h"
#include "keywords.h"
#include "draw.h"
#include "util.h"
#include "style.h"
#include "arg_t.h"

#include "mpm.h"

MBTK_INLINE void clrmap_append(unsigned long *clrmap, mbtk_kw_t kw, unsigned long clr)
{
	int n;
	for(n = 2; n < clrmap[0]+2; n+=2)
		if (clrmap[n] == kw)
			return;
	clrmap[n] = kw;
	clrmap[n+1] = clr;
	clrmap[0] += 2;
}

MBTK_INLINE unsigned long clr_lookup(const mbtk_style_t *style, mbtk_styleprop_t *sp, mbtk_kw_t kw, mbtk_color_t *dst)
{
	mbtk_arg_t arg;

	if ((sp != NULL) && (mbtk_style_get_prop(style, sp, kw, &arg) == 0) && (arg.type == MBTK_ULONG)) {
		if (dst != NULL) {
			dst->r = (arg.v.ul & 0xff0000000ul) >> 24;
			dst->g = (arg.v.ul & 0x000ff0000ul) >> 16;
			dst->b = (arg.v.ul & 0x00000ff00ul) >> 8;
			dst->a = (arg.v.ul & 0x0000000fful);
		}
		return arg.v.ul;
	}

	/* else return a stock invalid color */
	if (dst != NULL) {
		dst->r = dst->g = 255;
		dst->b = 0;
		dst->a = 255;
	}
	return 0xFFFF00FF;
}

static int mbtk_draw_import_anypm(mbtk_display_t *display, const mbtk_style_t *style, mbtk_styleprop_t *sp, mbtk_pixmap_t *dst, unsigned long *dst_clrmap, const char **mpm, int is_xpm)
{
	const char **line;
	int w, h, c, n, x, cpp, rv;
	mbtk_arg_t arg;
	mbtk_color_t clr[128] = {0};
	unsigned char *px, *pxs;


	if ((display == NULL) || (display->be == NULL) || (display->be->draw_import_pixmap == NULL))
		return -1;

	line = mpm;
	if (is_xpm) {
		if (strncmp(*line, "! XPM", 5) == 0)
			line++;
	}
	else {
		if (strcmp(*line, "! MPM1") != 0)
			return -1;
		line++;
	}

	if (is_xpm) {
		if (sscanf(*line, "%d %d %d %d", &w, &h, &c, &cpp) != 4)
			return -1;
		if (cpp != 1)
			return -1;
	}
	else {
		if (sscanf(*line, "%d %d %d", &w, &h, &c) != 3)
			return -1;
	}
	line++;

	if (dst_clrmap != NULL)
		dst_clrmap[0] = 0;

	/* read colors */
	for(n = 0; n < c; n++) {
		const char *args, *s = *line;
		int idx = s[0], len;
		unsigned long packed;
		mbtk_kw_t kw;

		if ((idx < 1) || (idx > 127))
			return -1; /* invalid color index */
		s++;
		if (!isspace(*s))
			return -1; /* invalid color: missing sep */
		while(isspace(*s)) s++;

		if (is_xpm) {
			if (*s != 'c')
				return -1; /* require 'c' for color */
			s++;
			while(isspace(*s)) s++;
		}

		if (strcmp(s, "None") == 0) {
			clr[idx].r = 0;
			clr[idx].g = 0;
			clr[idx].b = 0;
			clr[idx].a = 0;
		}
		else switch(*s) {
			case '#':
				arg.type = MBTK_STRING;
				arg.v.s = s;
				if (mbtk_arg_set_color(&(clr[idx]), &arg) != 0)
					return -1;
				break;
			case '$':
				if (is_xpm)
					return -1;
				s++;
				args = strpbrk(s, " \t");
				if (args != NULL) {
					char key[128];
					len = args - s;
					if (len >= sizeof(key))
						return -1; /* color property key too long */
					memcpy(key, s, len);
					key[len] = '\0';
					kw = mbtk_kw(key);
				}
				else
					kw = mbtk_kw(s);

				packed = clr_lookup(style, sp, kw, &clr[idx]);


				if (dst_clrmap != NULL)
					clrmap_append(dst_clrmap, kw, packed);

				/* parse optional $color args, e.g. alpha * 0.5 */
				if ((args != NULL) && (*args != '\0')) {
					while(isspace(*args)) args++;
					if (strncmp(args, "alpha", 5) == 0) {
						double alpha = clr[idx].a, d;
						char op;
						args += 5;
						while(isspace(*args)) args++;
						op = *args;
						args++;
						d = strtod(args, NULL);
						switch(op) {
							case '*': alpha *= d; break;
							case '/': alpha /= d; break;
							case '+': alpha += d; break;
							case '-': alpha -= d; break;
							case '=': alpha = d; break;
							default:
								return -1; /* invalid op */
						}
						if (alpha < 0) alpha = 0;
						else if (alpha > 255) alpha = 255;
						clr[idx].a = alpha;
					}
					else
						return -1; /* invalid color arg */
				}
				break;
			default:
				return -1; /* invalid color format */
		}
		line++;
	}

	/* set up header */
	dst->w = w;
	dst->h = h;
	dst->has_alpha = 1;

	/* convert pixels using the color table */
	pxs = px = malloc(w*h*4);
	for(n = 0; n < h; n++) {
		const char *s = *line;
		for(x = 0; x < w; x++) {
			int ci = *s;
			if (*s == '\0')
				return -1; /* premature end of line */
			if ((ci < 1) || (ci > 127))
				return -1; /* invalid color index */
			px[0] = clr[ci].r;
			px[1] = clr[ci].g;
			px[2] = clr[ci].b;
			px[3] = clr[ci].a;
			px += 4;
			s++;
		}
		if (*s != '\0')
			return -1; /* line too long */
		line++;
	}

	rv = display->be->draw_import_pixmap(display, dst, pxs, w, h, 1);
	free(pxs);
	return rv;
}

int mbtk_draw_import_mpm(mbtk_display_t *display, const mbtk_style_t *style, mbtk_styleprop_t *sp, mbtk_pixmap_t *dst, unsigned long *dst_clrmap, const char **mpm)
{
	return mbtk_draw_import_anypm(display, style, sp, dst, dst_clrmap, mpm, 0);
}

int mbtk_draw_import_xpm(mbtk_display_t *display, mbtk_pixmap_t *dst, const char **xpm)
{
	return mbtk_draw_import_anypm(display, NULL, NULL, dst, NULL, xpm, 1);
}

/* Return true if a given render's colors are matching sp's colors.
   Implemented in inverse: return 0 on first mismatch */
MBTK_INLINE int mpm_palette_match(mbtk_mpm_render_t *r, const mbtk_style_t *style, mbtk_styleprop_t *sp)
{
	int n;
	unsigned long *p;

	for(n = 0, p = r->palette.kw_clr; n < r->palette.len; n++,p+=2) {
		mbtk_kw_t key = p[0];
		unsigned long mpm_clr = p[1];
		unsigned long style_clr = clr_lookup(style, sp, key, NULL);
		if (style_clr != mpm_clr)
			return 0;
	}

	return 1;
}

/* Allocates a new render and renders mpms's mpm to it (but doesn't append it) */
MBTK_INLINE mbtk_mpm_render_t *mpm_render(mbtk_mpm_styled_t *mpms, mbtk_display_t *display, const mbtk_style_t *style, mbtk_styleprop_t *sp)
{
	mbtk_mpm_render_t *r = malloc(sizeof(mbtk_mpm_render_t));
	unsigned long clrmap[128*2+2];
	int n, rv;

	if (mpms->is_xpm)
		rv = mbtk_draw_import_xpm(display, &r->pxm, mpms->mpm);
	else
		rv = mbtk_draw_import_mpm(display, style, sp, &r->pxm, clrmap, mpms->mpm);

	if (rv != 0) {
		free(r);
		return NULL;
	}

	if (!mpms->is_xpm) {
		r->palette.len = clrmap[0]/2;
		r->palette.kw_clr = malloc(sizeof(unsigned long) * clrmap[0]);
		for(n = 2; n < clrmap[0]+2; n++)
			r->palette.kw_clr[n-2] = clrmap[n];
	}
	else {
		r->palette.len = 0;
		r->palette.kw_clr = NULL;
	}

	r->next = NULL;
	return r;
}

static mbtk_pixmap_t *mbtk_mpm_get_in_display(mbtk_display_t *display, const mbtk_style_t *style, mbtk_styleprop_t *sp, mbtk_kw_t id)
{
	mbtk_mpm_styled_t *mpms;
	mbtk_mpm_render_t *r, *last = NULL;

	mpms = htip_get(&display->mpm, id);
	if (mpms == NULL)
		return NULL;

	/* return a palette-matching render */
	for(r = mpms->head; r != NULL; r = r->next) {
		last = r;
		if (mpm_palette_match(r, style, sp))
			return &r->pxm;
	}

	/* no match, render new */
	r = mpm_render(mpms, display, style, sp);
	if (r == NULL)
		return NULL;

 /* append so that head is the display-default render for fast lookup of the common case */
	if (mpms->head == NULL)
		mpms->head = r;
	else
		last->next = r;
	return &r->pxm;
}

mbtk_px_id_t mbtk_mpm_get(mbtk_window_t *win, const mbtk_style_t *style, mbtk_styleprop_t *sp, mbtk_kw_t id, mbtk_pixmap_t **pxm_out)
{
	mbtk_pixmap_t *pxm = mbtk_mpm_get_in_display(win->display, style, sp, id);
	mbtk_px_id_t pid;

	if (pxm_out != NULL)
		*pxm_out = pxm;

	if (pxm == NULL)
		return mbtk_invalid_px_id;

	/* The pxm pointer is persistent: once a mbm+style combo is rendered,
	   it is kept until the display is destroyed. It is safe to use it as hash
	   key */
	pid = htpxid_get(&win->binds, pxm);
	if ((pid.p != mbtk_invalid_px_id.p) || (pid.l != mbtk_invalid_px_id.l))
		return pid;

	pid = mbtk_draw_bind_pixmap(win, pxm);
	htpxid_set(&win->binds, pxm, pid);

	return pid;
}

mbtk_pixmap_t *mbtk_mpm_get_first(mbtk_display_t *display, mbtk_kw_t id)
{
	mbtk_mpm_styled_t *mpms;

	mpms = htip_get(&display->mpm, id);
	if (mpms == NULL)
		return NULL;

	/* if there's no render yet, render the display-style-default variant so
	   that we don't need to return NULL. The display-style-default will be
	   needed sooner or later and the caller is interested in metadata, not
	   pixels */
	if (mpms->head == NULL)
		mpms->head = mpm_render(mpms, display, display->style, display->styleprop);

	return &mpms->head->pxm;
}


void mbtk_mpm_init_display(mbtk_display_t *display)
{
	htip_init(&display->mpm, longhash, longkeyeq);
}

void mbtk_mpm_uninit_display(mbtk_display_t *display)
{
	htip_entry_t *e;
	for(e = htip_first(&display->mpm); e != NULL; e = htip_next(&display->mpm, e)) {
		mbtk_mpm_render_t *r, *next;
		mbtk_mpm_styled_t *mpms = e->value;

		for(r = mpms->head; r != NULL; r = next) {
			next = r->next;
			mbtk_draw_free_pixmap(display, &r->pxm);
			free(r->palette.kw_clr);
			free(r);
		}
		free(mpms);
	}
	htip_uninit(&display->mpm);
}


static int mbtk_anypm_set(mbtk_display_t *display, mbtk_kw_t id, const char **mpm, int is_xpm)
{
	mbtk_mpm_styled_t *mpms;

	mpms = htip_get(&display->mpm, id);
	if (mpms != NULL)
		return -1;

	mpms = malloc(sizeof(mbtk_mpm_styled_t));
	mpms->mpm = mpm;
	mpms->is_xpm = is_xpm;
	mpms->head = NULL;
	htip_set(&display->mpm, id, mpms);

	return 0;
}

int mbtk_mpm_set(mbtk_display_t *display, mbtk_kw_t id, const char **mpm)
{
	return mbtk_anypm_set(display, id, mpm, 0);
}

int mbtk_xpm_set(mbtk_display_t *display, mbtk_kw_t id, const char **xpm)
{
	return mbtk_anypm_set(display, id, xpm, 1);
}

