#include <stdlib.h>
#include <string.h>
#include <libmbtk/style.h>
#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/handler.h>
#include <libmbtk/widget/hand_root.h>
#include <libmbtk/callback_imp.h>

hvbox_align_t mbtk_default_align = HVBOX_JUSTIFY;
hvbox_span_t mbtk_default_span = HVBOX_MINIMAL;
int mbtk_default_spacing = 2;
int mbtk_default_padding = 2;

void mbtk_window_set_root_widget(mbtk_window_t *win, mbtk_widget_t *widget)
{
	if (win->root == NULL) {
		mbtk_widget_t *root = calloc(sizeof(mbtk_widget_t), 1);
		root->hvbox.type = HVBOX_ROOT;
		root->win = win;
		root->flags.dirty = 1;
		win->root = &root->hvbox;
	}

	if (win->root->ch_first != NULL)
		mbtk_widget_destroy((mbtk_widget_t *)win->root->ch_first);

	win->root->ch_first = win->root->ch_last = &widget->hvbox;
	widget->hvbox.parent = win->root;
	widget->win = win;
	widget->flags.dirty = 1;


	if (widget->handler == NULL)
		mbtk_handler_root_install(widget);

	/* mbtk_be_win2render_win() makes a difference when win is a subwin */
	mbtk_widget_set_win(widget, win);
	mbtk_widget_assign(widget, mbtk_be_win2render_win(win));
}

void mbtk_widget_minsize(mbtk_widget_t *widget, long *mx, long *my)
{
	const mbtk_style_t *style = mbtk_widget_style(widget);
	if (style->minsize(style, widget, mx, my) != 0) {
		int res = -1;
		if (widget->imp->fallback_minsize != NULL)
			res = widget->imp->fallback_minsize(widget, mx, my);
		if (res != 0)
			*mx = *my = 1;
	}

	if (widget->w_min_u > *mx) *mx = widget->w_min_u;
	if (widget->h_min_u > *my) *my = widget->h_min_u;
}

static void mbtk_widget_draw(mbtk_widget_t *widget, int overlay)
{
	const mbtk_style_t *style = mbtk_widget_style(widget);
	int (*draw_fnc)(const mbtk_style_t *style, mbtk_widget_t *w) = overlay ? style->draw_widget_overlay : style->draw_widget;

	if ((draw_fnc(style, widget) != 0) && (widget->imp != NULL)) {
		/* no style-specific draw function for widget, fall back to widget's
		   style-independent generic draw */
		void (*draw_fnc)(mbtk_widget_t *w) = overlay ? widget->imp->fallback_draw_overlay : widget->imp->fallback_draw;
		if (draw_fnc != NULL)
			draw_fnc(widget);
	}
}

void mbtk_widget_show(mbtk_widget_t *widget, mbtk_bool_t recurse)
{
	if (widget->hvbox.hidden) {
		widget->hvbox.hidden = 0;
		mbtk_win_dirty_layout(widget->win);
	}
	if (recurse) {
		hvbox_t *h;
		for(h = widget->hvbox.ch_first; h != NULL; h = h->next)
			mbtk_widget_show((mbtk_widget_t *)h, 1);
	}
}

void mbtk_widget_show_up(mbtk_widget_t *widget, mbtk_bool_t recurse)
{
	hvbox_t *h;

	mbtk_widget_show(widget, recurse);

	for(h = &widget->hvbox; h != NULL; h = h->parent)
		if (h->hidden)
			mbtk_widget_show((mbtk_widget_t *)h, 0);
}



void mbtk_widget_hide(mbtk_widget_t *widget, mbtk_bool_t recurse)
{
	if (!widget->hvbox.hidden) {
		widget->hvbox.hidden = 1;
		mbtk_win_dirty_layout(widget->win);
	}
	if (recurse) {
		hvbox_t *h;
		for(h = widget->hvbox.ch_first; h != NULL; h = h->next)
			mbtk_widget_hide((mbtk_widget_t *)h, 1);
	}
}

int mbtk_widget_is_hidden(mbtk_widget_t *widget)
{
	hvbox_t *h;
	for(h = &widget->hvbox; h != NULL; h = h->parent)
		if (h->hidden)
			return 1;
	return 0;
}

void mbtk_widget_draw_recurse(mbtk_window_t *win, hvbox_t *root, int overlay)
{
	hvbox_t *h;
	mbtk_widget_t *w = (mbtk_widget_t *)root;

	if (root->hidden)
		return;

	mbtk_widget_draw(w, overlay);
	if (overlay)
		w->flags.dirty = 0;

	if (root->type != HVBOX_WIDGET)
		for(h = root->ch_first; h != NULL; h = h->next)
			mbtk_widget_draw_recurse(win, h, overlay);
}

void mbtk_widget_minsize_recurse(mbtk_widget_t *w)
{
	long w_min, h_min;
	hvbox_t *h;

	mbtk_widget_minsize(w, &w_min, &h_min);
	w->hvbox.w_min = w_min;
	w->hvbox.h_min = h_min;

	for(h = w->hvbox.ch_first; h != NULL; h = h->next)
		mbtk_widget_minsize_recurse((mbtk_widget_t *)h);
}

typedef struct {
	mbtk_kw_t key;
	mbtk_arg_t val;
	char s[1]; /* overallocated to "strdup" the string value of val */
} dly_propset_t;

/* Delayed styleprop: when a display is assigned, merge dispaly defaults
   and props explicitly set on the widget */
static void apply_styleprop_dly(mbtk_widget_t *w)
{
	const mbtk_style_t *disp_style;
	long n;

	if ((w->styleprop_dly == NULL) || (w->win == NULL))
		return;

	/* w->styleprop has a few fields set by mbtk_widget_style_set_prop(), but
	   it never inherited display defaults because there was no display available
	   at the time. This means only those fields of the widget's prop are not
	   really set but saved in ->styleprop_dly and we are going to apply them now. */

	/* set up styleprop as usual, copy defaults from display */
	disp_style = w->win->display->style;
	if (w->styleprop == NULL) {
		w->style = disp_style;
		w->styleprop = malloc(disp_style->propsize);
		memcpy(w->styleprop, w->win->display->styleprop, disp_style->propsize);
	}

	/* apply delayed sets */
	for(n = 0; n < w->styleprop_dly->used; n++) {
		dly_propset_t *ps = w->styleprop_dly->array[n];
		mbtk_widget_style_set_prop(w, ps->key, ps->val);
		free(ps);
	}

	/* remove delayed style-propsets from the widget */
	vtp0_uninit(w->styleprop_dly);
	free(w->styleprop_dly);
	w->styleprop_dly = NULL;
}

static void free_styleprop_dly(mbtk_widget_t *w)
{
	int n;

	if (w->styleprop_dly == NULL)
		return;

	for(n = 0; n < w->styleprop_dly->used; n++)
		free(w->styleprop_dly->array[n]);

	/* remove delayed style-propsets from the widget */
	vtp0_uninit(w->styleprop_dly);
	free(w->styleprop_dly);
	w->styleprop_dly = NULL;
}

int mbtk_widget_style_set_prop(mbtk_widget_t *dst, mbtk_kw_t key, mbtk_arg_t value)
{
	const mbtk_style_t *disp_style = NULL;
	int res;

	/* if no display is known save all sets for delayed set after assing */
	if (dst->win == NULL) {
		dly_propset_t *ps;
		int slen = 0;

		if (dst->styleprop_dly == NULL) {
			dst->styleprop_dly = malloc(sizeof(vtp0_t));
			vtp0_init(dst->styleprop_dly);
		}

		if (((value.type & 0xFF) == MBTK_STRING) && (value.v.s != NULL))
			slen = strlen(value.v.s) + 1;

		ps = malloc(sizeof(dly_propset_t) + slen);
		ps->key = key;
		memcpy(&ps->val, &value, sizeof(mbtk_arg_t));
		if (slen > 0) {
			memcpy(ps->s, value.v.s, slen);
			ps->val.v.s = ps->s;
		}
		vtp0_append(dst->styleprop_dly, ps);

		return 0;
	}

		/* if display is known, copy that style, else prepare for delayed style sets */
	disp_style = dst->win->display->style;

	if (dst->styleprop == NULL) {
		dst->style = disp_style;
		dst->styleprop = malloc(disp_style->propsize);
		memcpy(dst->styleprop, dst->win->display->styleprop, disp_style->propsize);
	}

	res = mbtk_style_set_prop(disp_style, dst->styleprop, key, value);
	if (res == 0)
		mbtk_widget_changed(dst, 0);
	return res;
}

int mbtk_widget_style_get_prop(mbtk_widget_t *w, mbtk_kw_t key, mbtk_arg_t *dst)
{
	const mbtk_style_t *style;

	if (w->win == NULL)
		return -1;

	style = mbtk_widget_style(w);
	if (w->styleprop != NULL)
		return mbtk_style_get_prop(style, w->styleprop, key, dst);

	return mbtk_display_style_get_prop(w->win->display, key, dst);
}


void mbtk_widget_changed(mbtk_widget_t *w, int force_layout)
{
	long w_min, h_min;
	int fr_l = w->hvbox.fr_l, fr_r = w->hvbox.fr_r, fr_t = w->hvbox.fr_t, fr_b = w->hvbox.fr_b;

	if (w->win == NULL)
		return;

	mbtk_win_dirty_draw(w->win);

	/* recalculate minimum size because the frame affects it */
	mbtk_widget_minsize(w, &w_min, &h_min);
	if (!force_layout && (w->hvbox.w_min == w_min) && (w->hvbox.h_min == h_min) && (fr_l == w->hvbox.fr_l) && (fr_r == w->hvbox.fr_r) && (fr_t == w->hvbox.fr_t) && (fr_b == w->hvbox.fr_b))
		return;

	w->hvbox.w_min = w_min;
	w->hvbox.h_min = h_min;

	/* redo the layout with the new minsize */
	mbtk_win_dirty_layout(w->win);
}

void mbtk_widget_needs_redraw(mbtk_widget_t *w)
{
	mbtk_win_dirty_draw(w->win);
}


void mbtk_widget_del_handler(mbtk_widget_t *w)
{
	if ((w->handler != NULL) && (w->handler->widget_remove != NULL))
		w->handler->widget_remove(w);
	w->handler = NULL;
	w->handler_data = NULL;
}

static mbtk_kw_t kw_pre_destroy;

void mbtk_widget_callback_pre_destroy(mbtk_widget_t *w, mbtk_event_handled_t (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data)
{
	mbtk_kw_cache(kw_pre_destroy, "pre_destroy");
	mbtk_callback_set_noarg(w, kw_pre_destroy, cb, user_data);
}

/* Remove window focus in case it is in w's subtree */
static void mbtk_widget_remove_focus_recurse(mbtk_window_t *win, mbtk_widget_t *w)
{
	hvbox_t *h;
	if (win->focus == w) {
		win->focus = NULL;
		return;
	}
	for(h = w->hvbox.ch_first; h != NULL; h = h->next)
		mbtk_widget_remove_focus_recurse(win, (mbtk_widget_t *)h);
}

void mbtk_widget_destroy(mbtk_widget_t *w)
{
	mbtk_window_t *win = w->win;
	hvbox_t *h, *next;

	if (win != NULL)
		mbtk_widget_remove_focus_recurse(win, w);

	free_styleprop_dly(w);

	if (w->hvbox.parent != NULL)
		hvbox_unlink(&w->hvbox);
	mbtk_widget_unassign(w);

	w->win = NULL;
	if (win != NULL)
		mbtk_win_dirty_layout(win);

	w->destroying = 1;
	if (w->destroy_lock)
		return;

	/* first recursively destroy children widgets */
	for(h = w->hvbox.ch_first; h != NULL; h = next) {
		next = h->next;
		mbtk_widget_destroy((mbtk_widget_t *)h);
	}

	/* call user callback */
	mbtk_kw_cache(kw_pre_destroy, "pre_destroy");
	mbtk_callback_noarg(w, kw_pre_destroy);

	/* call implementation-specific */
	if ((w->imp != NULL) && (w->imp->imp_destroy != NULL))
		w->imp->imp_destroy(w);

	mbtk_callback_uninit(w);

	/* remove from win and handler */
	mbtk_widget_del_handler(w);

	if (w->styleprop != NULL) {
		free(w->styleprop);
		w->styleprop = NULL;
	}

	/* free local allocation */
	if (!w->flags.caller_alloced)
		free(w);
}

void mbtk_widget_destroy_lock_inc(mbtk_widget_t *w)
{
	w->destroy_lock++;
}

void mbtk_widget_destroy_lock_dec(mbtk_widget_t *w)
{
	assert(w->destroy_lock > 0);
	if (w->destroy_lock == 0) return; /* if assert() is disabled */

	w->destroy_lock--;
	if ((w->destroy_lock == 0) && w->destroying)
		mbtk_widget_destroy(w);
}

void mbtk_widget_unlink(mbtk_widget_t *w)
{
	if (w->hvbox.parent != NULL)
		hvbox_unlink(&w->hvbox);

	if (w->win != NULL) {
		mbtk_widget_remove_focus_recurse(w->win, w);
		mbtk_win_dirty_layout(w->win);
		mbtk_widget_unassign(w);
		mbtk_widget_set_win(w, NULL);
	}
}


int mbtk_widget_set_prop(mbtk_widget_t *w, mbtk_kw_t key, const mbtk_arg_t *newval)
{
	if (w->imp->set_propa == NULL)
		return -1;
	return w->imp->set_propa(w, key, newval, 1);
}

int mbtk_widget_set_propa(mbtk_widget_t *w, mbtk_kw_t key, const mbtk_arg_t *newval, int arrlen)
{
	if (w->imp->set_propa == NULL)
		return -1;
	return w->imp->set_propa(w, key, newval, arrlen);
}

int mbtk_widget_get_prop(mbtk_widget_t *w, mbtk_kw_t key, mbtk_arg_t *dst)
{
	if (w->imp->get_propa == NULL)
		return -1;
	return w->imp->get_propa(w, key, dst, 1);
}

int mbtk_widget_get_propa(mbtk_widget_t *w, mbtk_kw_t key, mbtk_arg_t *dst, int arrlen)
{
	if (w->imp->get_propa == NULL)
		return -1;
	return w->imp->get_propa(w, key, dst, arrlen);
}


void mbtk_widget_set_win(mbtk_widget_t *w, mbtk_window_t *win)
{
	hvbox_t *h;

	w->win = win;

	apply_styleprop_dly(w);

	for(h = w->hvbox.ch_first; h != NULL; h = h->next)
		mbtk_widget_set_win((mbtk_widget_t *)h, win);
}

void mbtk_widget_assign(mbtk_widget_t *w, mbtk_window_t *render_win)
{
	hvbox_t *h;

	if (w->render_win != render_win) {
		if (w->render_win != NULL)
			mbtk_widget_unassign(w);

		w->render_win = render_win;
		if (w->imp->post_assign != NULL)
			w->imp->post_assign(w);
	}

	for(h = w->hvbox.ch_first; h != NULL; h = h->next)
		mbtk_widget_assign((mbtk_widget_t *)h, render_win);

	mbtk_widget_changed(w, 0);
}



void mbtk_widget_unassign(mbtk_widget_t *w)
{
	hvbox_t *h;

	for(h = w->hvbox.ch_first; h != NULL; h = h->next)
		mbtk_widget_unassign((mbtk_widget_t *)h);

	if (w->render_win != NULL) {
		if (w->imp->pre_unassign != NULL)
			w->imp->pre_unassign(w);

		w->render_win = NULL;
	}
}


