#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "backend.h"
#include "draw.h"
#include "display.h"
#include "util.h"
#include "window.h"
#define MBTK_NO_INCLUDE_WIDGETS
#include "widget.h"
#include "widget_priv.h"
#include "htpxid.h"
#include "handler.h"
#include "callback_imp.h"

mbtk_window_t *mbtk_window_new(mbtk_display_t *disp, mbtk_window_type_t type, const char *title, long x, long y, long w, long h)
{
	mbtk_window_t *win;

	if ((disp == NULL) || (disp->be == NULL) || (disp->be->window_new == NULL))
		return NULL;

	win = disp->be->window_new(disp, type, title, x, y, w, h);

	if (win == NULL)
		return NULL;

	win->display = disp;

	mbtk_window_get_size_(win, &w, &h);
	win->w = w;
	win->h = h;

	win->min_w_last = win->min_h_last = -1;

	if (disp->topwin == NULL)
		disp->topwin = win;

	memset(&win->link, 0, sizeof(win->link));
	gdl_append(&disp->windows, win, link);

	/* resource lists */
	htpxid_init(&win->binds, htpxid_hash, htpxid_keyeq);
	memset(&win->handlers, 0, sizeof(win->handlers));

	return win;
}

mbtk_window_t *mbtk_subwindow_new(mbtk_display_t *disp, mbtk_widget_t *parent, long x, long y, long w, long h)
{
	mbtk_window_t *win;

	if ((disp == NULL) || (disp->be == NULL) || (disp->be->subwindow_new == NULL))
		return NULL;

	win = disp->be->subwindow_new(disp, parent, x, y, w, h);
	win->in_widget = parent;
	win->x = x; win->y = y;
	win->w = w; win->h = h;
	return win;
}

void mbtk_subwindow_assign(mbtk_window_t *subwin, mbtk_window_t *direct_parent)
{
	mbtk_display_t *disp = direct_parent->display;
	subwin->display = disp;

	if ((disp != NULL) && (disp->be != NULL) && (disp->be->subwindow_assign != NULL))
		disp->be->subwindow_assign(subwin, direct_parent);
}

void mbtk_subwindow_unassign(mbtk_window_t *subwin)
{
	mbtk_display_t *disp = subwin->display;
	if ((disp != NULL) && (disp->be != NULL) && (disp->be->subwindow_unassign != NULL))
		disp->be->subwindow_unassign(subwin);
	subwin->display = NULL;
}

void mbtk_subwindow_geometry(mbtk_window_t *subwin, long x, long y, long w, long h)
{
	mbtk_display_t *disp = subwin->display;

	if ((x == subwin->x) && (y == subwin->y) && (w == subwin->w) && (h == subwin->h))
		return; /* no change */

	subwin->x = x; subwin->y = y;
	subwin->w = w; subwin->h = h;

	if ((disp == NULL) || (disp->be == NULL) || (disp->be->subwindow_geometry == NULL))
		return;

	disp->be->subwindow_geometry(subwin, x, y, w, h);
}


void mbtk_subwindow_free(mbtk_window_t *subwin)
{
	/* unassign will free whatever the backend allocated fields but leaves
	   the subwin struct behind. We, the high level lib, didn't allocate any
	   more fields, so we only need to free the subwin struct itself */
	free(subwin);
}

static void mbtk_window_free(mbtk_window_t *win)
{
	htpxid_entry_t *e;
	mbtk_handler_t *hr;

	vtp0_uninit(&win->watch_close);

	gdl_remove(&win->display->windows, win, link);

	if (win->display->topwin == win)
		mbtk_display_set_topwin(win->display, gdl_first(&win->display->windows));

	if (win->root != NULL)
		mbtk_widget_destroy((mbtk_widget_t *)win->root);

	for(hr = gdl_first(&win->handlers); hr != NULL; hr = gdl_first(&win->handlers)) {
		if (hr->handler_destroy != NULL)
			hr->handler_destroy(hr);
		else
			gdl_remove(&win->handlers, hr, link);
	}

	for(e = htpxid_first(&win->binds); e != NULL; e = htpxid_next(&win->binds, e)) {
		mbtk_px_id_t pid;

		pid = e->value;
		mbtk_draw_unbind_pixmap(win, pid);
	}

	htpxid_uninit(&win->binds);

	free(win);
}

void mbtk_window_set_title(mbtk_window_t *win, const char *title)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_set_title == NULL))
		return;

	win->display->be->window_set_title(win, title);
}

void mbtk_window_raise(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_raise == NULL))
		return;

	win->display->be->window_raise(win);
}

void mbtk_window_hide(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_hide == NULL))
		return;

	win->display->be->window_hide(win);

}

void mbtk_window_show(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_show == NULL))
		return;

	win->display->be->window_show(win);
}


void mbtk_window_transient4(mbtk_window_t *win, mbtk_window_t *tr_for)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_transient4 == NULL))
		return;

	win->display->be->window_transient4(win, tr_for);
}

mbtk_window_t *mbtk_be_win2render_win(mbtk_window_t *win)
{
	if (win == NULL)
		return NULL;

	if ((win->display == NULL) || (win->display->be == NULL) || (win->display->be->win2render_win == NULL)) {
		assert(!"backend shall have win2render_win()");
		return win;
	}

	return win->display->be->win2render_win(win);
}

void mbtk_window_close(mbtk_window_t *win)
{
	long n;

	if (win == NULL)
		return;

	/* increase close watches */
	for(n = 0; n < win->watch_close.used; n++) {
		int *cnt = win->watch_close.array[n];
		if (cnt != NULL)
			(*cnt)++;
	}

	if (win->root != NULL) {
		/* set win to NULL just in case some widgets destroys are delayed: those
		   widgets will be effectively destroyed only after the window is destroyed
		   in which case widget->win references would be invalid */
		mbtk_widget_set_win((mbtk_widget_t *)win->root, NULL);
		mbtk_widget_destroy((mbtk_widget_t *)win->root);
		win->root = NULL;
	}

	if ((win->display != NULL) && (win->display->be != NULL) && (win->display->be->window_close != NULL))
		win->display->be->window_close(win);

	mbtk_window_free(win);
}


static void mbtk_window_set_min_size_(mbtk_window_t *win)
{
	long w = MBTK_MAX(win->min_w_user, win->min_w_layout);
	long h = MBTK_MAX(win->min_h_user, win->min_h_layout);

	if ((w == win->min_w_last) && (h == win->min_h_last))
		return;

	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_set_min_size == NULL))
		return;

	win->display->be->window_set_min_size(win, w, h);
	win->min_w_last = w;
	win->min_h_last = h;
	if ((win->w < w) || (win->h < h)) {
		long new_w = MBTK_MAX(win->w, w), new_h = MBTK_MAX(win->h, h);
		mbtk_window_set_size(win, new_w, new_h);
	}
}


void mbtk_window_set_min_size(mbtk_window_t *win, long w, long h)
{
	win->min_w_user = w;
	win->min_h_user = h;

	mbtk_window_set_min_size_(win);
}

void mbtk_window_set_min_size_layout(mbtk_window_t *win, long w, long h)
{
	win->min_w_layout = w;
	win->min_h_layout = h;

	mbtk_window_set_min_size_(win);
}


void mbtk_window_set_size(mbtk_window_t *win, long w, long h)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_set_size == NULL))
		return;

	if (w < win->min_w_last) w = win->min_w_last;
	if (h < win->min_h_last) h = win->min_h_last;
	win->display->be->window_set_size(win, w, h);

	/* TODO: Is there a guarantee that size got set? */
	win->root->w = win->w = w;
	win->root->h = win->h = h;
}

int mbtk_window_get_size_(mbtk_window_t *win, long *w, long *h)
{
	int res;

	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_get_size == NULL))
		return -1;

	res = win->display->be->window_get_size(win, w, h);
	if (res == 0) {
		win->w = *w;
		win->h = *h;
	}
	return res;
}


void mbtk_win_resized(mbtk_window_t *win, long w, long h)
{
	if (win->root == NULL)
		return;
	win->root->w = w;
	win->root->h = h;
	win->w = w;
	win->h = h;
	mbtk_win_dirty_layout(win);
}

void mbtk_win_moved(mbtk_window_t *win, long x, long y)
{
	win->x = x;
	win->y = y;
}

void mbtk_win_redraw(mbtk_window_t *win)
{
	if (win->root == NULL) {
		win->layout_dirty = win->draw_dirty = 0;
		return;
	}

	if ((win->root->w == 0) || (win->root->h == 0)) {
		long wl, hl;
		if (mbtk_window_get_size_(win, &wl, &hl) != 0) {
			wl = win->w;
			hl = win->h;
		}
		win->layout_dirty = 1;
		win->root->w = wl;
		win->root->h = hl;
	}


	if (win->layout_dirty) {
		long prev_w = win->w, prev_h = win->h;

		/* calculate once for the minimum sizes */
		hvbox_update(win->root);
		mbtk_window_set_min_size_layout(win, win->root->w_min_calc, win->root->h_min_calc);

		/* if the window was smaller and had to resize, recalculate for the new size */
		if ((win->w != prev_w) || (win->h != prev_h)) {
			printf("window got resized to minimum\n");
			hvbox_update(win->root);
		}

		win->layout_dirty = 0;
		win->draw_dirty = 1;
/* hvbox_dump_nodes(stdout, 1, win->root, 0); */
	}


	if (win->draw_dirty) {
		/* first redraw every widget recursively then draw the overlays */
		mbtk_draw_start(win);
		mbtk_widget_draw_recurse(win, win->root, 0);
		mbtk_widget_draw_recurse(win, win->root, 1);
		mbtk_draw_flush(win);
		win->draw_dirty = 0;
	}
}

void mbtk_window_set_focus(mbtk_window_t *win, mbtk_widget_t *w, mbtk_bool_t w_has_pointer)
{
	if (win->focus != w) {
		if (win->focus != NULL) {
			win->focus->flags.focus = 0;
			mbtk_widget_changed(win->focus, 0);
		}
		win->focus = w;
		win->focus_has_pointer = w_has_pointer;
		w->flags.focus = 1;
		mbtk_widget_changed(w, 0);
	}
}


void mbtk_window_add_handler(mbtk_window_t *win, mbtk_handler_t *hr)
{
	if (win != NULL)
		gdl_append(&win->handlers, hr, link);
}

void mbtk_window_del_handler(mbtk_handler_t *hr)
{
	mbtk_window_t *win;

	if (hr->link.parent == NULL)
		return;

	win = (mbtk_window_t *)((char *)hr->link.parent - offsetof(mbtk_window_t, handlers));
	gdl_remove(&win->handlers, hr, link);
}

int mbtk_window_get_last_mouse_coords(mbtk_window_t *win, long *x, long *y)
{
	*x = win->mouse.last_x;
	*y = win->mouse.last_y;
	return win->entered;
}

void mbtk_window_add_watch_close(mbtk_window_t *window, int *dst)
{
	long n;
	for(n = 0; n < window->watch_close.used; n++)
		if (window->watch_close.array[n] == dst)
			return; /* already added */
	vtp0_append(&window->watch_close, dst);
}

void mbtk_window_del_watch_close(mbtk_window_t *window, int *dst)
{
	long n;
	for(n = 0; n < window->watch_close.used; n++) {
		if (window->watch_close.array[n] == dst) {
			vtp0_remove(&window->watch_close, n, 1);
			n--;
		}
	}
}

void mbtk_window_callback_close(mbtk_window_t *win, mbtk_event_handled_t (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data)
{
	mbtk_kw_t kw_win_close;
	mbtk_widget_t *w = mbtk_window_get_root(win);

	mbtk_kw_cache(kw_win_close, "win_close");
	mbtk_callback_set_noarg(w, kw_win_close, cb, user_data);
}

