#include <stdio.h>
#include <string.h>
#include "arg.h"
#include "log.h"
#include "dep.h"
#include "db.h"
#include "libs.h"
#include "src/tmpfawk/tmpfawk.h"
#include "Rev.h"
#include "../util/arg_auto_set.h"

#define version "0.1.1"


const arg_auto_set_t disable_libs[] = { /* list of --disable-LIBs and the subtree they affect */
	{"enable-glfw",       "/local/mbtk/want_glfw",      arg_true,      "$enable glfw (opengl) backend"},
	{"disable-glfw",      "/local/mbtk/want_glfw",      arg_false,     "$disable glfw (opengl) backend"},
	{"enable-sdl2",       "/local/mbtk/want_sdl2",      arg_true,      "$enable libsdl2 backend"},
	{"disable-sdl2",      "/local/mbtk/want_sdl2",      arg_false,     "$disable libsdl2 backend"},
	{"enable-xlib",       "/local/mbtk/want_xlib",      arg_true,      "$enable raw xlib (X11) backend"},
	{"disable-xlib",      "/local/mbtk/want_xlib",      arg_false,     "$disable raw xlib (X11) backend"},
	{"enable-test",       "/local/mbtk/want_test",      arg_true,      "$enable test backend"},
	{"disable-test",      "/local/mbtk/want_test",      arg_false,     "$disable test backend"},

	{NULL, NULL, NULL, NULL}
};

static void help1(void)
{
	printf(" --symbols                      enable symbols\n");
	printf(" --debug                        enable debug (no optimization, symbols)\n");
	printf(" --profile                      enable code profiling\n");
}

/* Runs when a custom command line argument is found
 returns true if no furhter argument processing should be done */
int hook_custom_arg(const char *key, const char *value)
{
	if (strcmp(key, "debug") == 0) {
		put("/local/mbtk/debug", strue);
		return 1;
	}
	if (strcmp(key, "symbols") == 0) {
		put("/local/mbtk/symbols", strue);
		return 1;
	}
	if (strcmp(key, "profile") == 0) {
		put("/local/mbtk/profile", strue);
		return 1;
	}

	if (arg_auto_set(key, value, disable_libs) == 0) {
		fprintf(stderr, "Error: unknown argument %s\n", key);
		exit(1);
	}
	return 1; /* handled by arg_auto_set() */
}


/* Runs before anything else */
int hook_preinit()
{
	return 0;
}

/* Runs after initialization */
int hook_postinit()
{
	db_mkdir("/local");
	db_mkdir("/local/mbtk");

	/* defaults */
	put("/local/mbtk/debug", sfalse);
	put("/local/mbtk/symbols", sfalse);
	put("/local/mbtk/profile", sfalse);
	put("/local/mbtk/prefix", "/usr/local");

	put("/local/mbtk/want_glfw", strue);
	put("/local/mbtk/want_sdl2", strue);
	put("/local/mbtk/want_xlib", strue);
	put("/local/mbtk/want_test", strue);

	return 0;
}

/* Runs after all arguments are read and parsed */
int hook_postarg()
{
	return 0;
}

/* Runs when things should be detected for the host system */
int hook_detect_host()
{
	return 0;
}

/* returns 1 if pthreads is available */
static int have_thread_lib_pthread(void)
{
	if (require("thread/pthread_create/*", 0, 0) != 0)
		return 0;
	if (require("libs/lpthread", 0, 0) != 0)
		return 0;
	if (require("thread/semget/*", 0, 0) != 0)
		return 0;
	put("/local/mbtk/thread/pthread", strue);
	return 1;
}

/* returns 1 if win32 threading is available */
static int have_thread_lib_win32(void)
{
	if (require("thread/CreateThread/*", 0, 0) != 0)
		return 0;
	if (require("thread/CreateSemaphore/*", 0, 0) != 0)
		return 0;
	put("/local/mbtk/thread/win32", strue);
	return 1;
}


/* returns 1 if there's threading available */
static int have_thread_lib(void)
{
	return have_thread_lib_pthread() || have_thread_lib_win32();
}

/* returns 1 pipe(2) is available */
static int have_pipe_unix(void)
{
	if (require("libs/io/pipe/*", 0, 0) != 0)
		return 0;
	return 1;
}

/* returns 1 if win32 threading is available */
static int have_pipe_win32(void)
{
	if (require("libs/io/_pipe/*", 0, 0) != 0)
		return 0;
	return 1;
}


/* returns 1 if there's any pipe() implementation is available */
static int have_pipe(void)
{
	return have_pipe_unix() || have_pipe_win32();
}

/* returns 1 if there's any way to get sub-second time from the system */
static int have_time_subsec(void)
{
	/* tests are done in order of preference (for monotonic time) */

	if (require("libs/time/clock_gettime/*", 0, 0) == 0) {
		require("libs/time/CLOCK_MONOTONIC_RAW/*", 0, 0); /* on Linux */
		return 1;
	}
	if (require("libs/time/mach_absolute_time/*", 0, 0) == 0) /* macos */
		return 1;
	if (require("libs/time/timeGetTime/*", 0, 0) == 0) /* winRT; prefer this over QPC below because it's faster */
		return 1;
	if (require("libs/time/QueryPerformanceCounter/*", 0, 0) == 0) /* win32 */
		return 1;
	if (require("libs/time/gettimeofday/*", 0, 0) == 0)
		return 1;
	if (require("libs/time/ftime/*", 0, 0) == 0)
		return 1;

	return 0;
}


/* Runs when things should be detected for the target system */
int hook_detect_target()
{
	const char *target_ansi, *target_ped, *target_pg, *target_no_pie, *target_c89, *target_c99, *target_wall, *cflg;
	int need_inl = 0;

	require("cc/cc", 0, 1);
	require("cc/fpic",  0, 0);
	require("cc/inline",  0, 0);
	require("cc/func_attr/unused/presents",  0, 0);
	require("cc/argstd/ansi",  0, 0);
	require("cc/argstd/pedantic",  0, 0);
	require("cc/argstd/pg",  0, 0);
	require("cc/argstd/no-pie",  0, 0);

	if (istrue(get("/local/mbtk/debug"))) {
		require("cc/argstd/Wall",  0, 0);
		require("cc/argstd/std_c89",  0, 0);
		require("cc/argstd/std_c99",  0, 0);
	}

	require("libs/socket/select/*", 0, 1); /* all current backends depend on this */

	target_ansi = get("cc/argstd/ansi");
	target_ped = get("cc/argstd/pedantic");
	target_pg = get("cc/argstd/pg");
	target_no_pie = get("cc/argstd/no-pie");
	target_wall = get("cc/argstd/Wall");
	target_c89 = get("cc/argstd/std_c89");
	target_c99 = get("cc/argstd/std_c99");


	put("/local/mbtk/cflags", "");

	/* common cflags */
	if (istrue(get("/local/mbtk/debug"))) {
		if (target_wall != NULL) {
			append("/local/mbtk/cflags", " ");
			append("/local/mbtk/cflags", target_wall);
		}
		append("/local/mbtk/cflags", " -g");
	}
	else
		append("/local/mbtk/cflags", " -O3");

	if (istrue(get("/local/mbtk/symbols")))
		append("/local/mbtk/cflags", " -g");

	if (istrue(get("/local/mbtk/profile"))) {
		append("/local/mbtk/cflags", " ");
		append("/local/mbtk/cflags", target_pg);
		append("/local/mbtk/cflags", " ");
		append("/local/mbtk/cflags", target_no_pie);
	}


	/* copy common and append restrictions for c89 and c99 */
	cflg = get("/local/mbtk/cflags");
	put("/local/mbtk/c89flags", cflg);
	put("/local/mbtk/c99flags", cflg);
	if (istrue(get("/local/mbtk/debug"))) {
		if (target_c89 != NULL) {
			append("/local/mbtk/c89flags", " ");
			append("/local/mbtk/c89flags", target_c89);
		}
		else if (target_ansi != NULL) {
			append("/local/mbtk/c89flags", " ");
			append("/local/mbtk/c89flags", target_ansi);
		}
		if (target_c99 != NULL) {
			append("/local/mbtk/c99flags", " ");
			append("/local/mbtk/c99flags", target_c99);
		}
		else if (target_ansi != NULL) {
			append("/local/mbtk/c99flags", " ");
			append("/local/mbtk/c99flags", target_ansi);
		}
		if (target_ped != NULL) {
			append("/local/mbtk/c89flags", " ");
			append("/local/mbtk/c89flags", target_ped);
			append("/local/mbtk/c99flags", " ");
			append("/local/mbtk/c99flags", target_ped);
		}
	}

	/* c89 doesn't have inline */
	if (!istrue(get("cc/inline")))
		need_inl = 1;

	if (need_inl) {
		/* disable inline for C89 */
		append("/local/mbtk/c89flags", " ");
		append("/local/mbtk/c89flags", "-Dinline= ");
	}

	/* detect libs needed for common code */
	require("libs/snprintf", 0, 0);

	/* detect inline after fixing up the flags */
	put("/local/mbtk/have_glfw", sfalse);
	put("/local/mbtk/have_sdl2", sfalse);
	put("/local/mbtk/have_xlib", sfalse);
	put("/local/mbtk/have_test", sfalse);

	if (istrue(get("/local/mbtk/want_glfw"))) {
		if (require("libs/gui/gl/*", 0, 0) != 0)
			report("gl not found, disabling glfw backend...");
		else if (require("libs/gui/glfw/*", 0, 0) != 0)
			report("glfw not found, disabling glfw backend...");
		else if (require("libs/sul/freetype2/*", 0, 0) != 0)
			report("freetype2 not found, disabling glfw backend...\n");
		else if (!have_thread_lib())
			report("threading not possible, disabling glfw...\n");
		else if (!have_pipe())
			report("no support for pipe, disabling glfw...\n");
		else
			put("/local/mbtk/have_glfw", strue);
	}

	if (istrue(get("/local/mbtk/want_sdl2"))) {
		if (require("libs/gui/sdl2/*", 0, 0) != 0)
			report("sdl2 (core lib, libsdl2) not found, disabling sdl2 backend...\n");
		else if (require("libs/gui/sdl2_gfx/*", 0, 0) != 0)
			report("sdl2_gfx not found, disabling sdl2 backend...\n");
		else if (require("libs/gui/sdl2_ttf/*", 0, 0) != 0)
			report("sdl2_ttf not found, disabling sdl2 backend...\n");
		else if (require("libs/sul/freetype2/*", 0, 0) != 0)
			report("freetype2 not found, disabling sdl2 backend...\n");
		else if (!have_pipe())
			report("no support for pipe, disabling sdl2...\n");
		else {
			require("libs/gui/sdl2_GetTicks64/*", 0, 0); /* safer timer option */
			put("/local/mbtk/have_sdl2", strue);
		}
	}

	if (istrue(get("/local/mbtk/want_xlib"))) {
		if (require("libs/gui/xopendisplay/*", 0, 0) != 0)
			report("xopendisplay() not found, disabling xlib backend...");
		else if (!have_time_subsec())
			report("no support for reading sub-second time, disabling xlib...\n");
		else
			put("/local/mbtk/have_xlib", strue);
		require("libs/gui/keysymtoucs4/*", 0, 0);
	}

	if (istrue(get("/local/mbtk/want_test")))
			put("/local/mbtk/have_test", strue);

/*	require("sys/types/size/4_u_int", 0, 1);*/

	return 0;
}

static void print_sum_setting_or(const char *node, const char *desc, int or)
{
	const char *res, *state;
	state = get(node);
	if (or)
		res = "enabled (implicit)";
	else if (istrue(state))
		res = "enabled";
	else if (isfalse(state))
		res = "disabled";
	else
		res = "UNKNOWN - disabled?";
	printf("%-55s %s\n", desc, res);
}

static void print_sum_setting(const char *node, const char *desc)
{
	print_sum_setting_or(node, desc, 0);
}

static void print_sum_cfg_val(const char *node, const char *desc)
{
	const char *state = get(node);
	printf("%-55s %s\n", desc, state);
}


static const char *IS_OK(int *generr, int val)
{
	*generr |= val;
	if (val == 0) return "ok";
	return "ERROR";
}

/* Runs after detection hooks, should generate the output (Makefiles, etc.) */
int hook_generate()
{
	int generr = 0;
	char *tmp, *rev = "non-svn";

	tmp = svn_info(0, "..", "Revision:");
	if (tmp != NULL) {
		rev = str_concat("", "svn r", tmp, NULL);
		free(tmp);
	}

	logprintf(0, "scconfig generate version info: version='%s' rev='%s'\n", version, rev);
	put("/local/mbtk/revision", rev);
	put("/local/mbtk/version",  version);
	put("/local/mbtk/sccbox", "../../scconfig/sccbox");

	printf("Generating tests/Makefile.conf... %s\n",   IS_OK(&generr, tmpfawk_rel("../tests", "Makefile.conf.in", "Makefile.conf")));
	printf("Generating doc/examples/Makefile.conf... %s\n",   IS_OK(&generr, tmpfawk_rel("../doc/examples", "Makefile.conf.in", "Makefile.conf")));
	printf("Generating libmbtk/Makefile... %s\n",      IS_OK(&generr, tmpfawk_rel("../libmbtk", "Makefile.in", "Makefile")));
	printf("Generating libmbtk/config.h... %s\n",      IS_OK(&generr, tmpfawk_rel("../libmbtk", "config.h.in", "config.h")));
	printf("Generating backend/glfw/Makefile... %s\n", IS_OK(&generr, tmpfawk_rel("../libmbtk/backend/glfw", "Makefile.in", "Makefile")));
	printf("Generating backend/sdl2/Makefile... %s\n", IS_OK(&generr, tmpfawk_rel("../libmbtk/backend/sdl2", "Makefile.in", "Makefile")));
	printf("Generating backend/xlib/Makefile... %s\n", IS_OK(&generr, tmpfawk_rel("../libmbtk/backend/xlib", "Makefile.in", "Makefile")));
	printf("Generating backend/test/Makefile... %s\n", IS_OK(&generr, tmpfawk_rel("../libmbtk/backend/test", "Makefile.in", "Makefile")));
	printf("Generating backend/Makefile... %s\n",      IS_OK(&generr, tmpfawk_rel("../libmbtk/backend", "Makefile.in", "Makefile")));

	if (!generr) {
		printf("\n\n");
		printf("=====================\n");
		printf("Configuration summary\n");
		printf("=====================\n");

		print_sum_setting("/local/mbtk/debug",          "Compilation for debugging");
		print_sum_setting_or("/local/mbtk/symbols",     "Include debug symbols", istrue(get("/local/mbtk/debug")));
		print_sum_cfg_val("/local/mbtk/prefix",         "installation prefix (--prefix)");

		printf("\nBackends:\n\n");
		print_sum_setting("/local/mbtk/have_glfw",      "glfw (opengl)");
		print_sum_setting("/local/mbtk/have_sdl2",      "libsdl2");
		print_sum_setting("/local/mbtk/have_xlib",      "raw X11");
		print_sum_setting("/local/mbtk/have_test",      "test");

		if (!istrue(get("/local/mbtk/have_glfw")) && !istrue(get("/local/mbtk/have_sdl2")) && !istrue(get("/local/mbtk/have_xlib")) && !istrue(get("/local/mbtk/have_test"))) {
			printf("\nConfiguration FAILED: no backend configured.\n\n");
			generr = 1;
		}
		if (!istrue(get("/local/mbtk/have_glfw")) && !istrue(get("/local/mbtk/have_sdl2")) && !istrue(get("/local/mbtk/have_xlib"))) {
			printf("\nConfiguration WARNING: no graphical backend configured.\nGUI without graphic is not too useful.\n\n");
		}
		else {
			FILE *f = fopen("Rev.stamp", "w");
			fprintf(f, "%d", myrev);
			fclose(f);
			printf("\nConfiguration complete, ready to compile.\n\n");
		}
	}
	else
		fprintf(stderr, "\nError generating some of the files\n");

	return generr;
}

/* Runs before everything is uninitialized */
void hook_preuninit()
{
}

/* Runs at the very end, when everything is already uninitialized */
void hook_postuninit()
{
}

